#include <iostream>
#include <unistd.h>
#include "CcReceiver.h"

CcReceiver::CcReceiver(int Ca, int Vpid, int Apid[], int Dpid[])
:cReceiver(Ca, -1, Vpid, Apid, Dpid, NULL), cThread("CcReceiver")
{
  dprint(1, "CcReceiver::CcReceiver()");
  
  cMutexLock lock(&m_Mutex);
  m_parser = NULL;
  m_lastPacketSize = -1;
  m_queue_size = 0;
  errCnt = 0;
}
 
CcReceiver::~CcReceiver(void)
{
  dprint(1, "CcReceiver::~CcReceiver()");
  Detach();
  Cancel(0);
}

void CcReceiver::Action(void)
{
  int t0;
  byte data[READ_BUFFER];
  int len = 0;
  dprint(1, "CcReceiver::Action()");

  while (Running()) {
        t0 = time_ms();
        len = GetUserData((byte *)&data);
        
        if(len>0){		
        	dprint(3, "CcReceiver::Action() loop (%d bytes)", len);
        	while(m_parser->ParseBuffer(data, len));	
			m_parser->InitParser();
		}
		else
			if(errCnt++<3)
				dprint(3, "CcReceiver::Action(): NO USER DATA FOUND");
							
        cCondWait::SleepMs(RECEIVER_FREQ - (time_ms() - t0));
    }
}

void CcReceiver::setExtractor(CcExtractor *extractor){
	m_parser = extractor;
}

void CcReceiver::Activate(bool On)
{
  dprint(1, "CcReceiver::Activate()");
  if (On) {
	dprint(2, "CcReceiver::Activate() - activating...");
     	if (!Running())
         Start();
	else
	 dprint(2, "CcReceiver::Activate() - was already running.");
  }
  else{
	dprint(2, "CcReceiver::Activate() - deactivating...");
  	if (Running())
	 Cancel(0);
	else
	 dprint(2, "CcReceiver::Activate() - was not running.");
  }
}

CC_Data *CcReceiver::GetCC(){
	dprint(3,"CcReceiver::getCC\n");
	if(m_parser==NULL){
		dprint(3, "Queue is empty.\n");
		return NULL;
	}
	return m_parser->PopCC();
}

void CcReceiver::Receive(uchar *Data, int Length)
{
	if(m_lastPacketSize!=Length){
		dprint(1, "CC: Packet size: %d bytes\n", Length);
		m_lastPacketSize=Length;
	}
	if (Length == TS_SIZE) {
     /* the following originates from libdvbmpeg: */
     if (!(Data[3] & PAYLOAD)) {
        return;
        }
     uint8_t off = 0;
     if (Data[3] & ADAPT_FIELD) {
        off = Data[4] + 1;
        }
     if (Data[1] & PAY_START) {
        uint8_t *sb = Data + 4 + off;
        if (sb[7] & PTS_DTS_FLAGS) {
           uint8_t *pay = sb + sb[8] + 9; 
           int l = TS_SIZE - 13 - off - sb[8];
           if((m_parser != NULL) && (l>0)){
		dprint(3, "Data stored (%d/%d).\n", m_queue_size, Length);
		StoreStream(pay, l);					 }
	   }
     /* end */
     }
  }	
}

bool CcReceiver::StoreStream(byte *data, int len){
	dprint(4, "StoreStream\n");
	cMutexLock lock(&m_Mutex_Data);
	int size = len;
	
	if((len + m_queue_size)>=READ_BUFFER)
		size = READ_BUFFER - m_queue_size;		
	
	for(int i=0; i<size; i++)
		m_queue[m_queue_size++] = data[i];
		
	return (size == len);
}

int CcReceiver::GetUserData(byte *store){
	dprint(4, "CcReceiver::GetUserData()");
	cMutexLock lock(&m_Mutex_Data);
	dprint(4, "CcReceiver::GetUserData() - locked");
	int r = m_queue_size;
	for(int i=0; i< m_queue_size; i++)
		store[i] = m_queue[i];
	m_queue_size = 0;
	dprint(4, "CcReceiver::GetUserData() - finnished");
	return r;	
}
